import copy
import json

import pytest

from internetarchive.iarequest import (
    MetadataRequest,
    S3Request,
    prepare_files_patch,
    prepare_patch,
    prepare_target_patch,
)
from tests.conftest import PROTOCOL, IaRequestsMock


@pytest.fixture
def sample_metadata():
    return copy.deepcopy({
        "metadata": {"title": "Test"},
        "files": [
            {"name": "test.txt", "custom": {"tags": ["old"]}, "foo": "bar"},
        ],
        "dupe_pallet_index": {
            "IA9999": ["IA999901"]
        }
    })


@pytest.mark.parametrize(("metadata", "expected"), [
    ({"custom": ["new"]}, [{'op': 'add', 'path': '/custom', 'value': ['new']}]),
    ({"title": "New Title"}, [{'op': 'replace', 'path': '/title', 'value': 'New Title'}]),
    ({"title": "REMOVE_TAG"}, [{'op': 'remove', 'path': '/title'}]),
])
def test_metadata_patch_operations(metadata, expected, sample_metadata):
    patch = prepare_patch(
        metadata=metadata,
        source_metadata=sample_metadata["metadata"],
        append=False,
        append_list=False,
        insert=False,
    )
    assert patch == expected


@pytest.mark.parametrize(("metadata", "expected"), [
    ({"new-key": ["new", "new2"]}, [{'op': 'add', 'path': '/new-key', 'value': ['new', 'new2']}]),
    ({"custom": "foo new"}, [{'op': 'replace', 'path': '/custom', 'value': 'foo new'}]),
    ({"custom": "REMOVE_TAG"}, [{'op': 'remove', 'path': '/custom'}]),
])
def test_file_metadata_patch_operations(metadata, expected, sample_metadata):
    patch = prepare_files_patch(
        metadata=metadata,
        files_metadata=sample_metadata["files"],
        target="files/test.txt",
        append=False,
        append_list=False,
        insert=False,
        expect={}
    )
    assert patch == expected


@pytest.mark.parametrize(("metadata", "expected"), [
    (
        {"IA9999": ["UPDATED"], "NEW_ITEM": ["NEW123"]},
        [
            {'op': 'add', 'path': '/NEW_ITEM', 'value': ['NEW123']},
            {'op': 'replace', 'path': '/IA9999/0', 'value': 'UPDATED'}
        ]
    ),
])
def test_target_patch_add_and_replace(metadata, expected, sample_metadata):
    patch = prepare_target_patch(
        metadata=metadata,
        source_metadata=sample_metadata,
        target="dupe_pallet_index",
        append=False,
        append_list=False,
        insert=False,
        expect={}
    )
    assert patch == expected


@pytest.mark.parametrize(("metadata", "expected"), [
    (
        {"IA9999": ["IA999901", "IA999902", "IA999903"]},
        [{'op': 'add', 'path': '/IA9999/1', 'value': ['IA999901', 'IA999902', 'IA999903']}]
    ),
    (
        {"IA9999": "IA999902"},
        [{'op': 'add', 'path': '/IA9999/1', 'value': 'IA999902'}]
    ),
])
def test_target_patch_append_list(metadata, expected, sample_metadata):
    patch = prepare_target_patch(
        metadata=metadata,
        source_metadata=sample_metadata,
        target="dupe_pallet_index",
        append=False,
        append_list=True,
        insert=False,
        expect={}
    )
    assert patch == expected


def test_metadata_request_patch_key(sample_metadata):
    with IaRequestsMock() as rsps:
        rsps.add_metadata_mock('test_item', body=json.dumps(sample_metadata))

        req = MetadataRequest(
            metadata={"title": "New Title"},
            url=f"{PROTOCOL}//archive.org/metadata/test_item"
        )
        prepared = req.prepare()
        assert any(k.endswith('-patch') for k in prepared.data)


@pytest.mark.parametrize(("test_value", "expected"), [
    (
        "http://example.com/foo bar?q=✓",
        "uri(http%3A//example.com/foo%20bar%3Fq%3D%E2%9C%93)"
    ),
])
def test_metadata_header_uri_encoding(test_value, expected):
    req = S3Request(
        method='PUT',
        url=f"{PROTOCOL}//s3.us.archive.org/test_item",
        metadata={"source": test_value},
        access_key='test_access',
        secret_key='test_secret'
    )
    prepared = req.prepare()
    header = prepared.headers.get('x-archive-meta00-source', '')
    assert header == expected



#
# The internetarchive module is a Python/CLI interface to Archive.org.
#
# Copyright (C) 2012-2024 Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
internetarchive.exceptions
~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (C) 2012-2024 by Internet Archive.
:license: AGPL 3, see LICENSE for more details.
"""

from typing import Dict, Optional


class AuthenticationError(Exception):
    """Authentication Failed"""


class ItemLocateError(Exception):
    def __init__(self, *args, **kwargs):
        default_message = "Item cannot be located because it is dark or does not exist."
        if args or kwargs:
            super().__init__(*args, **kwargs)
        else:
            super().__init__(default_message)


class InvalidChecksumError(Exception):
    def __init__(self, *args, **kwargs):
        default_message = "File corrupt, checksums do not match."
        if args or kwargs:
            super().__init__(*args, **kwargs)
        else:
            super().__init__(default_message)


class AccountAPIError(Exception):
    """Base exception for Account API-related errors."""
    def __init__(self, message: str, error_data: Optional[Dict] = None):
        super().__init__(message)
        self.error_data = error_data
